package dk.dtu.s094758.lib.processor;

import dk.dtu.s094758.lib.exception.ClassificationException;
import dk.dtu.s094758.lib.credentials.ClassificationLevel;
import dk.dtu.s094758.lib.credentials.AccessCredentials;
import dk.dtu.s094758.lib.annotation.ClassifiedWebMethod;
import dk.dtu.s094758.lib.annotation.ClassifiedWebService;
import dk.dtu.s094758.lib.model.ClassifiedVariable;
import dk.dtu.s094758.lib.ws.SupportsClassification;
import java.io.IOException;
import java.io.Writer;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Generated;
import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.Filer;
import javax.annotation.processing.Messager;
import javax.annotation.processing.ProcessingEnvironment;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.jws.WebMethod;
import javax.jws.WebService;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.Name;
import javax.lang.model.element.PackageElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.TypeMirror;
import javax.lang.model.util.ElementFilter;
import javax.lang.model.util.Elements;
import javax.tools.JavaFileObject;

/**
 *
 * @author linas
 */
@SupportedAnnotationTypes(value = {"dk.dtu.s094758.lib.annotation.ClassifiedWebService"})
@SupportedSourceVersion(SourceVersion.RELEASE_6)
public class WebServiceClassificationProcessor extends AbstractProcessor {

    private Filer filer;
    private Messager messager;
    private Elements elementUtils;

    @Override
    public void init(ProcessingEnvironment processingEnv) {
        super.init(processingEnv);
        filer = processingEnv.getFiler();
        messager = processingEnv.getMessager();
        elementUtils = processingEnv.getElementUtils();
    }

    @Override
    public boolean process(Set<? extends TypeElement> annotations, RoundEnvironment roundEnvironment) {

        System.out.println("Processing classification annotations...");

        Set<? extends Element> annotatedWebServices = roundEnvironment.getElementsAnnotatedWith(ClassifiedWebService.class);

        for (Element annotatedWebService : annotatedWebServices) {

            PackageElement packageElement = elementUtils.getPackageOf(annotatedWebService);

            Name packageName = packageElement.getQualifiedName();
            String classifiedPackageName = packageName + ".classified";

            Name className = annotatedWebService.getSimpleName();
            Element fullClassName = annotatedWebService;

            JavaFileObject javaFile = null;
            try {
                javaFile = filer.createSourceFile(classifiedPackageName + "." + className);
            } catch (IOException ex) {
                Logger.getLogger(WebServiceClassificationProcessor.class.getName()).log(Level.SEVERE, null, ex);
            }

            Writer writer = null;
            try {
                writer = javaFile.openWriter();

                writer.write("package " + classifiedPackageName + ";\n");
                writer.write("\n");

                ClassifiedWebService classifiedWebService = annotatedWebService.getAnnotation(ClassifiedWebService.class);
                String targetNamespace = classifiedWebService.targetNamespace();
                String serviceName = classifiedWebService.serviceName();
                String portName = classifiedWebService.portName();
                String wsdlLocation = classifiedWebService.wsdlLocation();

                writer.write("@" + Generated.class.getName() + "(\"" + packageName + "." + className + "\")\n");
                writer.write("@" + SupportsClassification.class.getName() + "()\n");
                writer.write("@" + WebService.class.getName() + "(targetNamespace = \"" + targetNamespace + "\", serviceName = \"" + serviceName + "\", portName = \"" + portName + "\", wsdlLocation = \"" + wsdlLocation + "\")\n");
                writer.write("public class " + className + " extends " + fullClassName + " {\n");
                writer.write("\n");

                for (ExecutableElement method : ElementFilter.methodsIn(annotatedWebService.getEnclosedElements())) {

                    Name methodName = method.getSimpleName();
                    TypeMirror returnType = method.getReturnType();

                    String modifiersString = buildModifiersString(method.getModifiers());
                    String methodParametersString = buildMethodParametersString(method.getParameters());
                    String callParametersString = buildCallParametersString(method.getParameters());

                    ClassifiedWebMethod classifiedWebMethod = method.getAnnotation(ClassifiedWebMethod.class);
                    String operationName = classifiedWebMethod.operationName();
                    String action = classifiedWebMethod.action();
                    Boolean exclude = classifiedWebMethod.exclude();
                    ClassificationLevel classificationLevel = classifiedWebMethod.classificationLevel();

                    writer.write("    @" + Generated.class.getName() + "(\"" + packageName + "." + className + "." + methodName + "()\")\n");
                    writer.write("    @" + WebMethod.class.getName() + "(operationName = \"" + operationName + "\", action = \"" + action + "\", exclude = " + exclude + ")\n");

                    writer.write("    " + modifiersString + " " + returnType + " " + methodName + "(" + methodParametersString + ") throws " + ClassificationException.class.getName() + " {\n");
                    writer.write("\n");

                    writer.write("        " + ClassificationLevel.class.getName() + " classificationLevel = accessCredentials.getClassificationLevel();\n");
                    writer.write("\n");

                    writer.write("        if (classificationLevel.getNumericValue() < " + ClassificationLevel.class.getName() + "." + classificationLevel.toString() + ".getNumericValue()) {\n");
                    writer.write("            throw new " + ClassificationException.class.getName() + "();\n");
                    writer.write("        }\n");
                    writer.write("\n");

                    writer.write("        " + returnType + " returnData = super." + methodName + "(" + callParametersString + ");\n");
                    writer.write("\n");

                    writer.write("        Class<? extends " + returnType + "> returnClass = returnData.getClass();\n");
                    writer.write("        " + Method.class.getName() + "[] methods = returnClass.getMethods();\n");
                    writer.write("        " + Map.class.getName() + "<" + String.class.getName() + ", " + ClassifiedVariable.class.getName() + "> classifiedVariableMap = " + WebServiceClassificationUtils.class.getName() + ".buildClassifiedVariableMap(methods);\n");
                    writer.write("\n");
                    writer.write("        for (" + Iterator.class.getName() + "<" + Entry.class.getCanonicalName() + "<" + String.class.getName() + ", " + ClassifiedVariable.class.getName() + ">> iterator = classifiedVariableMap.entrySet().iterator(); iterator.hasNext();) {\n");
                    writer.write("\n");
                    writer.write("            " + Entry.class.getCanonicalName() + "<" + String.class.getName() + ", " + ClassifiedVariable.class.getName() + "> entry = iterator.next();\n");
                    writer.write("            " + ClassifiedVariable.class.getName() + " classifiedVariable = entry.getValue();\n");
                    writer.write("\n");
                    writer.write("            if (classifiedVariable.getClassifiedGetter().getClassificationLevel().getNumericValue()\n");
                    writer.write("                    > classifiedVariable.getClassifiedSetter().getClassificationLevel().getNumericValue()) {\n");
                    writer.write("\n");
                    writer.write("                throw new " + ClassificationException.class.getName() + "();\n");
                    writer.write("            }\n");
                    writer.write("\n");
                    writer.write("            if (classifiedVariable.getClassifiedGetter().getClassificationLevel().getNumericValue()\n");
                    writer.write("                    > " + ClassificationLevel.class.getName() + "." + classificationLevel.toString() + ".getNumericValue()) {\n");
                    writer.write("\n");
                    writer.write("                " + Method.class.getName() + " method = classifiedVariable.getClassifiedSetter().getMethod();\n");
                    writer.write("                " + Object.class.getName() + " nullObject = null;\n");
                    writer.write("\n");
                    writer.write("                try {\n");
                    writer.write("                    method.invoke(returnData, nullObject);\n");
                    writer.write("                } catch (" + IllegalAccessException.class.getName() + " ex) {\n");
                    writer.write("                    throw new " + ClassificationException.class.getName() + "();\n");
                    writer.write("                } catch (" + IllegalArgumentException.class.getName() + " ex) {\n");
                    writer.write("                    throw new " + ClassificationException.class.getName() + "();\n");
                    writer.write("                } catch (" + InvocationTargetException.class.getName() + " ex) {\n");
                    writer.write("                    throw new " + ClassificationException.class.getName() + "();\n");
                    writer.write("                }\n");
                    writer.write("            }\n");
                    writer.write("        }\n");
                    writer.write("\n");

                    writer.write("        return returnData;\n");
                    writer.write("    }\n");
                }

                writer.write("}\n");

                writer.flush();
                writer.close();

            } catch (IOException ex) {
                Logger.getLogger(WebServiceClassificationProcessor.class.getName()).log(Level.SEVERE, null, ex);
            }
        }

        return true;
    }

    private String buildModifiersString(Set<Modifier> modifiers) {

        StringBuilder modifierStringBuilder = new StringBuilder();

        if (!modifiers.isEmpty()) {

            for (Modifier modifier : modifiers) {
                modifierStringBuilder.append(modifier.toString());
                modifierStringBuilder.append(" ");
            }

            modifierStringBuilder.deleteCharAt(modifierStringBuilder.length() - 1);
        }

        return modifierStringBuilder.toString();
    }

    private String buildMethodParametersString(List<? extends VariableElement> parameters) {

        StringBuilder parameterStringBuilder = new StringBuilder();

        if (!parameters.isEmpty()) {

            for (VariableElement parameter : parameters) {

                String parameterName = parameter.getSimpleName().toString();
                String parameterType = parameter.asType().toString();

                parameterStringBuilder.append(parameterType);
                parameterStringBuilder.append(" ");
                parameterStringBuilder.append(parameterName);
                parameterStringBuilder.append(", ");
            }
        }

        parameterStringBuilder.append(AccessCredentials.class.getName());
        parameterStringBuilder.append(" ");
        parameterStringBuilder.append("accessCredentials");

        return parameterStringBuilder.toString();
    }

    private String buildCallParametersString(List<? extends VariableElement> parameters) {

        StringBuilder parameterStringBuilder = new StringBuilder();

        if (!parameters.isEmpty()) {

            for (VariableElement parameter : parameters) {

                String parameterName = parameter.getSimpleName().toString();

                parameterStringBuilder.append(parameterName);
                parameterStringBuilder.append(", ");
            }

            parameterStringBuilder.delete(parameterStringBuilder.length() - 2, parameterStringBuilder.length());
        }

        return parameterStringBuilder.toString();
    }
}
