package dk.dtu.s094758.generator;

import java.lang.reflect.Method;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.MatchResult;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.xmlbeans.SchemaAnnotation;
import org.apache.xmlbeans.SchemaLocalElement;
import org.apache.xmlbeans.SchemaParticle;
import org.apache.xmlbeans.SchemaType;
import org.apache.xmlbeans.XmlObject;

/**
 *
 * @author linas
 */
public class Generator {

    private static final Pattern labelPattern = Pattern.compile("<label>(.*)</label>");

    public static String generateServiceClass(String serviceName, String packageName, String... methodStrings) {

        StringBuilder stringBuilder = new StringBuilder();

        stringBuilder.append("package ");
        stringBuilder.append(packageName);
        stringBuilder.append(";\n\npublic class ");
        stringBuilder.append(serviceName);
        stringBuilder.append(" {\n\n");

        for (String methodString : methodStrings) {
            stringBuilder.append(methodString);
        }

        stringBuilder.append("}\n");

        return stringBuilder.toString();
    }

    public static String generateServiceMethod(Class<?> classType, SchemaType schemaType,
            Class<?> documentClassType, SchemaType documentSchemaType,
            Class<?> responseClassType, SchemaType responseSchemaType) {

        StringBuilder stringBuilder = new StringBuilder();

        String className = classType.getSimpleName();
        String methodName = className.substring(0, 1).toLowerCase() + className.substring(1);

        Method returnMethod = null;

        try {
            returnMethod = responseClassType.getDeclaredMethod("getReturn");
        } catch (NoSuchMethodException ex) {
            Logger.getLogger(Generator.class.getName()).log(Level.SEVERE, null, ex);
        } catch (SecurityException ex) {
            Logger.getLogger(Generator.class.getName()).log(Level.SEVERE, null, ex);
        }

        Class<?> returnType = returnMethod.getReturnType();
        String returnTypeName = returnType.getCanonicalName();

        SchemaAnnotation methodAnnotation = ((SchemaLocalElement) documentSchemaType.getContentModel()).getAnnotation();
        String methodLabel = getJifLabel(methodAnnotation);

        stringBuilder.append("\tpublic ");
        stringBuilder.append(returnTypeName);
        stringBuilder.append("{");
        stringBuilder.append(methodLabel);
        stringBuilder.append("} ");
        stringBuilder.append(methodName);
        stringBuilder.append("(");

        Method[] methods = classType.getDeclaredMethods();
        SchemaParticle[] particleChildren = schemaType.getContentModel().getParticleChildren();

        if (particleChildren != null) {

            for (SchemaParticle schemaParticle : particleChildren) {

                String name = schemaParticle.getName().toString();
                String type = getTypeOfVar(name, methods);

                SchemaAnnotation annotation = ((SchemaLocalElement) schemaParticle).getAnnotation();
                String jifLabel = getJifLabel(annotation);

                stringBuilder.append(type);
                stringBuilder.append("{");
                stringBuilder.append(jifLabel);
                stringBuilder.append("} ");
                stringBuilder.append(name);
                stringBuilder.append(", ");
            }
        } else {

            String name = schemaType.getContentModel().getName().toString();
            String type = getTypeOfVar(name, methods);

            SchemaAnnotation annotation = ((SchemaLocalElement) schemaType.getContentModel()).getAnnotation();
            String jifLabel = getJifLabel(annotation);

            stringBuilder.append(type);
            stringBuilder.append("{");
            stringBuilder.append(jifLabel);
            stringBuilder.append("} ");
            stringBuilder.append(name);
            stringBuilder.append(", ");
        }

        stringBuilder.delete(stringBuilder.length() - 2, stringBuilder.length());
        // FIXME: this is a total hack that only supports boolean.
        stringBuilder.append(") {\n\t\t return true;\n\t}\n");

        return stringBuilder.toString();
    }

    public static String generateBeanClass(Class<?> classType, SchemaType schemaType) {

        StringBuilder stringBuilder = new StringBuilder();

        String packageName = classType.getPackage().getName();
        String className = classType.getSimpleName();

        stringBuilder.append("package ");
        stringBuilder.append(packageName);
        stringBuilder.append(";\n\npublic class ");
        stringBuilder.append(className);
        stringBuilder.append(" {\n\n");

        Method[] methods = classType.getDeclaredMethods();
        SchemaParticle[] particleChildren = schemaType.getContentModel().getParticleChildren();

        for (SchemaParticle schemaParticle : particleChildren) {

            String name = schemaParticle.getName().toString();
            String type = getTypeOfVar(name, methods);

            SchemaAnnotation annotation = ((SchemaLocalElement) schemaParticle).getAnnotation();
            String capName = name.substring(0, 1).toUpperCase() + name.substring(1);
            String jifLabel = getJifLabel(annotation);

            stringBuilder.append("\tprivate ");
            stringBuilder.append(type);
            stringBuilder.append("{");
            stringBuilder.append(jifLabel);
            stringBuilder.append("} ");
            stringBuilder.append(name);
            stringBuilder.append(";\n\n\tpublic ");
            stringBuilder.append(type);
            stringBuilder.append("{");
            stringBuilder.append(jifLabel);
            stringBuilder.append("} get");
            stringBuilder.append(capName);
            stringBuilder.append("() {\n");
            stringBuilder.append("\t\treturn ");
            stringBuilder.append(name);
            stringBuilder.append(";\n\t}\n\n\tpublic void set");
            stringBuilder.append(capName);
            stringBuilder.append("{");
            stringBuilder.append(jifLabel);
            stringBuilder.append("}(");
            stringBuilder.append(type);
            stringBuilder.append("{");
            stringBuilder.append(jifLabel);
            stringBuilder.append("} ");
            stringBuilder.append(name);
            stringBuilder.append(") {\n\t\tthis.");
            stringBuilder.append(name);
            stringBuilder.append(" = ");
            stringBuilder.append(name);
            stringBuilder.append(";\n\t}\n\n");
        }

        stringBuilder.append("}\n");

        return stringBuilder.toString();
    }

    private static String getTypeOfVar(String varName, Method[] methods) {

        String type = null;

        for (Method method : methods) {

            String methodName = method.getName();

            if (methodName.equalsIgnoreCase("get" + varName)) {
                type = method.getReturnType().getCanonicalName();
            }
        }

        return type;
    }

    private static String getJifLabel(SchemaAnnotation annotation) {

        String jifLabel = null;

        XmlObject[] applicationInformation = annotation.getApplicationInformation();

        for (XmlObject xmlObject : applicationInformation) {

            String xmlText = xmlObject.xmlText();
            Matcher matcher = labelPattern.matcher(xmlText);

            if (matcher.find()) {
                MatchResult matchResult = matcher.toMatchResult();
                String match = matchResult.group(1);
                jifLabel = match.replaceFirst("<label>", "").replaceFirst("</label>", "").replaceAll("top", "*").replaceAll("bottom", "_").replaceAll("writer", "<-").replaceAll("reader", "->");
            }
        }

        return jifLabel;
    }
}
